def test_expected_output(testdir):
    string = """
    Lorem ipsum
    ```python
    print(1 + 3)
    print(1 - 3)
    print(1 * 3)
    ```
    dolor sit amet
    <!--pytest-codeblocks:expected-output-->
    ```
    4
    -2
    3
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_expected_output_fail(testdir):
    string = """
    Lorem ipsum
    ```python
    print(1 + 3)
    ```
    dolor sit amet
    <!--pytest-codeblocks:expected-output-->
    ```
    5
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(failed=1)



def test_shell(testdir):
    string = """
    Lorem ipsum
    ```sh
    ls
    ```
    dolor sit amet
    ```sh
    cd
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=2)


def test_shell_fail(testdir):
    string = """
    ```sh
    cdc
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(failed=1)


def test_shell_expect_fail(testdir):
    string = """
    <!--pytest-codeblocks:expect-error-->
    ```sh
    cdc
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_shell_expect_fail_passed(testdir):
    string = """
    <!--pytest-codeblocks:expect-error-->
    ```sh
    cd
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(failed=1)


def test_shell_expect_output(testdir):
    string = """
    ```sh
    echo abc
    ```
    <!--pytest-codeblocks:expected-output-->
    ```sh
    abc
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_shell_expect_output_fail(testdir):
    string = """
    ```sh
    echo abc
    ```
    <!--pytest-codeblocks:expected-output-->
    ```sh
    ac
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(failed=1)


def test_bash(testdir):
    string = """
    ```bash
    foo=1
    if [[ $foo == 1 ]]; then
        echo abc
    fi
    ```
    <!--pytest-codeblocks:expected-output-->
    ```sh
    abc
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)



import io

import pytest

import pytest_codeblocks


def test_unclosed(testdir):
    string = """
    ```python
    1 + 2 + 3
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(errors=1)


def test_maxlines():
    unclosed = io.StringIO(
        """
    ```python
    1 + 2 + 3
    ```
    """
    )
    with pytest.raises(RuntimeError):
        pytest_codeblocks.extract_from_buffer(unclosed, max_num_lines=1)



import contextlib
import re
import sys

# namedtuple with default arguments
# <https://stackoverflow.com/a/18348004/353337>
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from typing import Optional, Union


@dataclass
class CodeBlock:
    code: str
    lineno: int
    syntax: Optional[str] = None
    expected_output: Optional[str] = None
    expect_exception: bool = False


def extract_from_file(
    f: Union[str, bytes, Path], encoding: Optional[str] = None, *args, **kwargs
):
    with open(f, encoding=encoding) as handle:
        return extract_from_buffer(handle, *args, **kwargs)


def extract_from_buffer(f, max_num_lines: int = 10000):
    out = []
    previous_nonempty_line = None
    k = 1

    while True:
        line = f.readline()
        k += 1
        if not line:
            # EOF
            break

        if line.strip() == "":
            continue

        if line.lstrip()[:3] == "```":
            syntax = line.strip()[3:]
            num_leading_spaces = len(line) - len(line.lstrip())
            lineno = k - 1
            # read the block
            code_block = []
            while True:
                line = f.readline()
                k += 1
                if not line:
                    raise RuntimeError("Hit end-of-file prematurely. Syntax error?")
                if k > max_num_lines:
                    raise RuntimeError(
                        f"File too large (> {max_num_lines} lines). Set max_num_lines."
                    )
                # check if end of block
                if line.lstrip()[:3] == "```":
                    break
                # Cut (at most) num_leading_spaces leading spaces
                nls = min(num_leading_spaces, len(line) - len(line.lstrip()))
                line = line[nls:]
                code_block.append(line)

            if previous_nonempty_line is None:
                out.append(CodeBlock("".join(code_block), lineno, syntax))
                continue

            # check for keywords
            m = re.match(
                "<!--pytest-codeblocks:(.*)-->", previous_nonempty_line.strip()
            )
            if m is None:
                out.append(CodeBlock("".join(code_block), lineno, syntax))
                continue

            keyword = m.group(1)

            # handle special tags
            if keyword == "skip":
                continue
            elif keyword == "expected-output":
                if len(out) == 0:
                    raise RuntimeError(
                        "Found <!--pytest-codeblocks-expected-output--> "
                        + "but no previous code block."
                    )
                if out[-1].expected_output is not None:
                    raise RuntimeError(
                        "Found <!--pytest-codeblocks-expected-output--> "
                        + "but block already has expected_output."
                    )
                expected_output = "".join(code_block)
                out[-1] = CodeBlock(
                    out[-1].code, out[-1].lineno, out[-1].syntax, expected_output
                )
            elif keyword == "cont":
                if len(out) == 0:
                    raise RuntimeError(
                        "Found <!--pytest-codeblocks-cont--> but no previous code block."
                    )
                out[-1] = CodeBlock(
                    out[-1].code + "".join(code_block),
                    out[-1].lineno,
                    out[-1].syntax,
                    out[-1].expected_output,
                    out[-1].expect_exception,
                )
            elif keyword in ["expect-exception", "expect-error"]:
                out.append(
                    CodeBlock(
                        "".join(code_block), lineno, syntax, expect_exception=True
                    )
                )
            else:
                raise RuntimeError('Unknown pytest-codeblocks keyword "{keyword}."')

        previous_nonempty_line = line

    return out


# https://stackoverflow.com/a/3906390/353337
@contextlib.contextmanager
def stdout_io(stdout=None):
    old = sys.stdout
    if stdout is None:
        stdout = StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old



def test_expect_error(testdir):
    string = """
    Lorem ipsum
    <!--pytest-codeblocks:expect-exception-->
    ```python
    raise RuntimeError()
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_expect_error_fail(testdir):
    string1 = """
    Lorem ipsum
    <!--pytest-codeblocks:expect-exception-->
    ```python
    1 + 1
    ```
    """
    testdir.makefile(".md", string1)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(failed=1)



def test_cont(testdir):
    string = """
    Lorem ipsum
    ```python
    a = 1
    ```
    dolor sit amet
    <!--pytest-codeblocks:cont-->
    ```
    a + 1
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_hidden_cont(testdir):
    string = """
    Lorem ipsum
    <!--
    ```python
    a = 1
    ```
    -->
    dolor sit amet
    <!--pytest-codeblocks:cont-->
    ```
    a + 1
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_nocont(testdir):
    string = """
    <!--pytest-codeblocks:cont-->
    ```python
    1 + 2 + 3
    ```
    """
    testdir.makefile(".md", string)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(errors=1)



#
# Take a look at the example
# https://docs.pytest.org/en/stable/example/nonpython.html
#
import subprocess

import pytest

from .main import extract_from_file, stdout_io


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group.addoption(
        "--codeblocks", action="store_true", help="enable testing of codeblocks"
    )


def pytest_collect_file(path, parent):
    config = parent.config
    if config.option.codeblocks and path.ext == ".md":
        return MarkdownFile.from_parent(parent, fspath=path)


class MarkdownFile(pytest.File):
    def __init__(self, fspath, parent):
        super().__init__(fspath, parent)

    def collect(self):
        for block in extract_from_file(self.fspath):
            if block.syntax not in ["python", "sh", "bash"]:
                continue
            # https://docs.pytest.org/en/stable/deprecations.html#node-construction-changed-to-node-from-parent
            out = Codeblock.from_parent(parent=self, name=self.name)
            out.obj = block
            yield out


class Codeblock(pytest.Item):
    def __init__(self, name, parent, obj=None):
        super().__init__(name, parent=parent)
        self.obj = obj

    # TODO for python 3.7+, stdout=subprocess.PIPE can be replaced by
    #      capture_output=True
    def runtest(self):
        output = None
        if self.obj.syntax == "python":
            if self.obj.expect_exception:
                with pytest.raises(Exception):
                    exec(self.obj.code, {"__MODULE__": "__main__"})
            else:
                with stdout_io() as s:
                    try:
                        # https://stackoverflow.com/a/62851176/353337
                        exec(self.obj.code, {"__MODULE__": "__main__"})
                    except Exception as e:
                        raise RuntimeError(
                            f"{self.name}, line {self.obj.lineno}:\n```\n"
                            + self.obj.code
                            + "```\n\n"
                            + f"{e}"
                        )
                output = s.getvalue()
        else:
            assert self.obj.syntax in ["sh", "bash"]
            executable = {
                "sh": None,
                "bash": "/bin/bash",
                "zsh": "/bin/zsh",
            }[self.obj.syntax]
            if self.obj.expect_exception:
                with pytest.raises(Exception):
                    subprocess.run(
                        self.obj.code, shell=True, check=True, executable=executable
                    )
            else:
                ret = subprocess.run(
                    self.obj.code,
                    shell=True,
                    check=True,
                    stdout=subprocess.PIPE,
                    executable=executable,
                )
                output = ret.stdout.decode()

        if output is not None and self.obj.expected_output is not None:
            if self.obj.expected_output != output:
                raise RuntimeError(
                    f"{self.name}, line {self.obj.lineno}:\n```\n"
                    + f"Expected output\n```\n{self.obj.expected_output}```\n"
                    + f"but got\n```\n{output}```"
                )

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        # if isinstance(excinfo.value, CodeblockException):
        return excinfo.value.args[0]
        # if excinfo.errisinstance(RuntimeError):
        #     return excinfo.value.args[0].stdout
        # return super().repr_failure(excinfo)

    def reportinfo(self):
        return (self.fspath, -1, "code block check")



import pathlib

import pytest_codeblocks


def test_basic(testdir):
    string1 = """
    Lorem ipsum
    ```python
    1 + 2 + 3
    ```
    dolor sit amet
    """
    testdir.makefile(".md", string1)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=1)


def test_skip(testdir):
    string1 = """
    Lorem ipsum
    <!--pytest-codeblocks:skip-->
    ```python
    1 + 2 + 3
    ```
    dolor sit amet

    Some newlines:

    <!--pytest-codeblocks:skip-->

    ```python
    1 + 2 + 3
    ```
    """
    testdir.makefile(".md", string1)
    result = testdir.runpytest("--codeblocks")
    result.assert_outcomes(passed=0)


# example.md against reference strings test against
def test_reference():
    ref = [
        pytest_codeblocks.CodeBlock("1 + 1\n", 1, "python"),
        pytest_codeblocks.CodeBlock("1 + 2 + 3\n2 + 5\n", 5, "python"),
        pytest_codeblocks.CodeBlock(
            "import pytest_codeblocks\n\npytest_codeblocks.extract_from_buffer\n",
            10,
            "python",
        ),
        pytest_codeblocks.CodeBlock("foobar\n", 16, "ruby"),
        pytest_codeblocks.CodeBlock("echo abc\n", 20, "sh"),
        pytest_codeblocks.CodeBlock("# ```import math```\n", 30, "python"),
        pytest_codeblocks.CodeBlock("1 + 1 == 2\n", 35, "python"),
        pytest_codeblocks.CodeBlock("1 + 1 == 2\n", 40, "python"),
    ]
    this_dir = pathlib.Path(__file__).resolve().parent
    lst = pytest_codeblocks.extract_from_file(this_dir / "example.md")
    print(lst)
    for r, obj in zip(ref, lst):
        print("r  ", r)
        print("obj", obj)
        assert r == obj
